\ Short piece of code from the nRF24 mesh network implementation

: WRITE-MODE    ( -- )          \ Power up module as transmitter
    08 29 *bic  wakeup          \ P2OUT  CE low, receive off, wakeup transmitter
    reset  1 pipes-on  2 /ms ;  \ Reset flags & pipe-0 active, wait 200 microsec.

: READ-MODE     ( -- )
    0F 00 write-reg  2 pipes-on \ Power up module as receiver, pipe-1 active
    reset  08 29 *bis  2 /ms ;  \ P2OUT  Enable receive mode, wait 200 microsec.

\ Send command & receive data while doing maximal x 'retries'
\ The result flag is false when all is ok.
: TRANSCEIVE    ( cT retries -- flag )
    0 ?do
        red-on  write-mode      \ Led on, go transmitting
        dup xemit?  0A < if     \ Send char command
            reset  read-mode    \ Reset flags, go receiving
            read? if            \ Data received?
                drop  xkey drop \ Read payload, success
                08 029 *bic  false \ P2OUT  CE low, receive mode off
                red-off  unloop exit \ Ready!
            then
        then
    loop
    drop  red-off  08 29 *bic   \ P2OUT  CE low, receive mode off
    true ;                      \ Transceive failed somewhere!

\ Set destination address to node from stack, receive address is #me "me"
: SET-DEST      ( node -- )
    dup to dest  1 >pay  #me 2 >pay \ Set Destination & origin nodes
    F0 F0 F0 F0 dest 0A write-addr  \ Receive address P0
    F0 F0 F0 F0 dest 10 write-addr ; \ Transmit address

\ Add a direct found node to my node tables
: NOTE-DIRECT   ( node -- )
    >r  r@ all get* 0= if \ Yes, node not recorded yet?
        r@ new *set \ Yes, note new node number!
    then
    r@ r@ nodes +cc \ Add direct node to table
    r@ direct *set  \ Save that node number
    r> all *set ;   \ Note node in ALL nodes table too

\ 0 scan = max. 4 meters  ( 1 wall )    2 scan = max. 10 meters ( 1 wall )
\ 4 scan = max. 10 meters ( 2 walls )   6 scan = max. .. meters ( .. )
: SCANX)    ( rf-setting -- ) \ Scan to other nodes within RF-range
    >nrf  new zero          \ Set RF-power, empty new node table
     nodes c@ 0= if         \ Not yet scanned?
        #me all *set        \ Yes, i am there too
        #n 0 ?do            \ Try all findable nodes
            i #me <> if     \ Do not scan yourself
                i set-dest
                ch S retry transceive 0= if  \ Node found?
                    tcr  i t.       \ Yes. show the scanned node number
                    t" Found" i note-direct
                then
            then
        loop
    then
    cr ." New: " new .map  nrf >nrf \ Copy ALL table & show
    copyall  .direct  called  read-mode ; \ show direct nodes found

: SCANX     ( -- )      4 scanx) ;  \ SCANX at medium power
