(*     **** New plan for mesh network implementation ****

        Basic 24l01 using usci 310319:  1722 bytes
        Mesh node 3.15:                 4282 bytes
        Total:                          6004 bytes
        Free:                           1324 bytes

Payload format seventeen bytes now:
|  0  |  1  |  2  |   3    |   4   |  5 to 11   |
|-----|-----|-----|--------|-------|------------|
|Admin|Dest.|Orig.|Sub node|Command|d00| to |d0C|

0 pay>       = Administration (Auxiliary) byte
1 pay>       = Destination node
2 pay>       = Origin node
3 pay>       = Address of sub node at destination
4 pay>       = Command for destination node
5 to 11 pay> = Data 0C bytes

*)

hex
\ Defining this node
1 constant #TYPE        \ I am a switch
0 constant #SUB         \ I contain +n sub nodes
10 constant #N  \ Max. number of nodes used
#n 8 / constant #MAP \ Size of bitmap

\ EXEC handles execution tables, it needs the table address on the
\ stack. Note that, all data needs te be cell aligned. The last command
\ token is always -1 and ERR-XT then is the error recovery routine.
\ Like: CREATE EXEC-TABLE  C0 ,  XT0 ,  C1 , XT1  ...  -1 , ERR-XT ,
\ c = command token, exec-table = start address of execution table
code EXEC   ( c exec-table -- ?? )
4708 ,  4437 ,  482A ,  5228 ,  933A ,  2402 ,
970A ,  23FA ,  8328 ,  4437 ,  4826 ,  4630 ,
end-code
code RESTORE-SP     4214 ,  adr s0 ,  4F00 ,  end-code
code LOC    ( node a -- bit byte-adr ) \ bit location in byte-addr
  4428 ,  F038 ,  7 ,  431A ,  8318 ,  2802 , 5A0A ,
  3FFC ,  4428 ,  4A84 ,  0 ,  422A ,  831A ,  2402 ,
  1108 ,  3FFC ,  F038 ,  #map 1- ,  5807 ,  4F00 ,
end-code
\ : LOC   ( node a -- bit byte-adr ) \ bit location in byte-addr
\    >r  dup 7 and           \ node bit-nr
\    1 swap lshift           \ node bit
\    swap 3 rshift           \ node/8
\    [ #map 1- ] literal and \ Mask excess bits
\    r> + ;                  \ Leave bit & byte-adr


\ Node data tables
create NEW      #map allot  \ New found nodes
create NESTED   #map allot  \ New found nested nodes
create USE      #map allot  \ Scratch bit table
create ALL      #map allot  \ BIT-table with all found nodes
create DIRECT   #map allot  \ BIT-table with direct accessable nodes
create INDIRECT #map allot  \ BIT-table with indirect accessable nodes
create ABO      #map allot  \ BIT-table for subscription data
create NODES #n 2* 1+ allot \ Table with all accessable nodes
create TYPES    #n allot    \ Table with node-types

: ZERO  ( a -- )        #map 0 fill ; \ Erase bit-map a
: *SET  ( node a -- )   loc *bis ; \ Add node to a
: *CLR  ( node a -- )   loc *bic ; \ Remove node from a
: GET*  ( node a -- b ) loc bit* ; \ Node present in a?

\ Leave number of nodes found in a bitmap
: COUNT*    ( a -- +n ) \ Counted noted nodes
    0  #n 0 ?do
        over i swap get* \ Node present?
        if   1+  then   \ Add 1 when found
    loop  nip ;

\ Leave node number of the first used node in bitmap & erase it
: NEXT?     ( a -- false | node true )
    #n 0 ?do            \ Test all bits
        i over get*     \ Node bit set?
        if  i swap *clr  i true unloop exit  then
    loop  drop  false ;  \ Nothing found


\ Basic control values
value END?      \ -1 when close token is received (-1=ready, 0=running)
value CNT       \ Command/Ping counter
value ORG?      \ Is -1 for starting node only
value ORG       \ Remember node to answer too
value TK        \ Remember command token
value RETRY     \ XEMIT retry counter
0C constant #B  \ Databuffer size in payload

\ Node primitives XEMIT now with retry!! Possibly with restart after every 2 attempts
: XEMIT     ( c -- )
    red-on  write-mode
    retry 0 ?do
        dup temit  dup xemit? 0A <> if \ Send payload, show echo
            drop  red-off  read-mode  unloop exit
        then
    loop
    drop  red-off  setup24l01  read-mode ; \ TX failed

: C+!       ( n a -- )  tuck c@ +  swap c! ; \ Increase byte address
: +CC       ( c c a -- ) >r  r@ count 2* + >r  r@ c!  r> 1+ c!  1 r> c+! ;
: CCRANGE   ( a -- a2 a1 ) count 2* bounds ; \ Convert addr len to addr range?
: ?TIMEOUT  ( f -- )    if  A0  else  300  then  to delay ;
: NOTDONE?  ( n -- f )  indirect get* 0= ;  \ Indirect mode not wired?
: >USE      ( a -- )    use #map move ;     \ Copy new found nodes
: >PAYLOAD  ( a -- )    6 '>pay #map move ; \ Copy map a to payload
\ Switch node states
: READY     ( -- )      -1 to end? ; \ External command ready
: START     ( -- )      0 to end? ;  \ External command called
: HALT      ( -- )      -1 to org? ; \ Stop free running programs
: FREE      ( -- )      0 to org? ;  \ Release ORG? flag

\ Aternative MS routine
value MS?   \ Restart MS when false
value MS)   \ Decreases 976 times each second
code INT-ON   ( -- )    D232 ,  4F00 ,  end-code
\ Clock = 8000000/8192 longest interval 67,10 sec. usable as MS
\ Decrease (ms) until it's zero
create MSTIMER  9382 ,  adr ms) ,  2402 ,  53B2 ,  adr ms) ,  1300 ,
: HALT-MS   ( -- )      5A91 120 !  0 to ms?  ready ; \ Stop MS timer
: >MS       ( u -- )    5A19 120 !  to ms)  -1 to ms?  start ; \ (Re)start MS timer
: MS-ON     ( -- )      5A91 120 !  1 0 *bis  0 to ms?  int-on ; \ IE1  Activate MS
: (MS)      ( -- )      ms? if  ms) if exit then  halt-ms   then ;  \ MS controller
mstimer   FFF4 vec!     \ Install watchdog interrupt vector


\ ADC on P1.0 and sample time at 64 clocks
: ADC-SETUP ( -- )
    2 1B0 **bic     \ ADC10CTL0  Clear ENC
    1 4A c!         \ ADC10AE0   P1.0 = ADC in
    1810 1B0 ! ;    \ ADC10CTL0  Sampletime 64 clocks, ADC on

\ We need to clear the ENC bit before setting a new input channel
: ADC       ( +n -- u )
    2 1B0 **bic     \ ADC10CTL0  Clear ENC
    F000 and 1B2 !  \ ADC10CTL1  Select input, MCLK/5
    3 1B0 **bis     \ ADC10CTL0  Set ENC & ADC10SC
    begin 1 1B2 bit** 0= until  \ ADC10CTL1  ADC10 busy?
    1B4 @ ;         \ ADC10MEM   Read result

: ADC-IN    ( -- +n )   0000 adc 2 rshift ; \ 8-bit ADC-value


\ PWM power output, period length is 1000 clock cycles ( #CYCLUS )
dm 1000 constant #CYCLUS
value PWM   \ Remember last given PWM value
: >POWER    ( 0 to #CYCLUS -- )
    dup 8 u< if  drop 0  then \ Values lower then 8 are zero
    #cyclus umin  196 !     \ TA1CCR2   Set pulselength
    010 029 *bic ;          \ P2OUT

\ Led PWM power driver at P2.4
: PWM-SETUP ( -- )
    10 02E *bis             \ P2SEL  Set PWM to output pin P2.4
    0 180 !   #cyclus 192 ! \ TA1CTL, TA1CCR0  Set period time
    40 186 !                \ TA1CCTL2 Set output mode positive pulse
    234  180 !              \ TA1CTL   Activate timer
    8 to pwm  pwm >power ;  \ Basic brightness


: >DEST     ( node -- )     \ Node is destination when no hop is used
    nodes ccrange ?do       \ Then check hopping nodes \ n
        i c@ over = if      \ Destination exists?      \ n
            i 1+ c@ dup set-dest \ Yes, send to "hopping" node \ n h
            over = ?timeout \ When true use short timeout \ n
            1 >pay  unloop exit \ add dest. node to payload & ready
        then
    2 +loop  set-dest  -1 ?timeout ; \ Node not in tables, just try!

: >NODE     ( c node -- )
   t? if  dup 1 .r  then  >dest xemit ; \ Send command c to node n

: RESET     ( -- )
( ) all zero  direct zero   \ No 'empty' nodes found
( ) indirect zero  abo zero  0 nodes c!
( ) types #n 0 fill  0 to cnt ; \ Empty type table, Reset ping

: RESTART   ( -- )
    spi-setup               \ SPI-active
    -1 to status            \ Init. status
    dm 100 to delay         \ Wait 11 ms for an answer after a command
    default-rf  setup24L01  \ Init. nRF24
    free  3 to retry        \ Three retries
    #me set-dest  ready     \ Init. addresses
    reset  .status  tron ;  \ Empty nodes table, print status, tracer on

\ Administration byte is used for the answer token
: RESPOND)  ( c node -- )   >r  0 >pay  ch } r> >node ; \ Send finish token
: RESPOND   ( c -- )        2 pay> respond) ;

\ Ask node for data, leaves false when a node address is not valid
: ASK?      ( c node -- flag )  \ Node is node address
    >r  r@ >dest  retry transceive 0= if \ Info received?
        r@ 2 pay> = if          \ Yes, is it from addressed node?
            r> drop  true  read-mode  exit \ Ready success!
        else
            tcr t" Wrong node " r@ t. 2 pay> t.
        then
    then
    r> drop  read-mode  false ;    \ Failed!

\ Print all noted nodes from a bitmap
: .MAP      ( a -- )
    dup count* 0= if  ." no "  then  ." nodes "
    #n 0 ?do
        i over get* if  i .  then \ Show nodes found
    loop
    ." present "  drop ;


\ The 'S' response routines for every node. Leave all relevant node data
: <SCANX    ( -- )
( ) ch S temit
    nodes c@ 5 >pay \ Node table allready filled?
    #type 6 >pay    \ Node type
    #sub 7 >pay     \ Number of sub nodes, 0 to 9
(   ...  )          \ 8 to 12, max. nine sub node types
    ch W respond ;  \ Vanity ;)

\ Return 'H' answer, the nr of direct nodes and the bitmap of these nodes
: <HOP      ( -- )
    ch H temit  direct >payload  direct count*  respond ;

\ Return 'N' answer, the number of indirect nodes and the bitmap of these nodes
: <HOPP     ( -- )
    ch N temit  indirect >payload  indirect count*  respond ;

\ Return 'P' answer to origin node with a 16-bit counter
: <PING     ( -- )
    cnt 5 >pay  cnt >< 6 >pay  ch P temit  incr cnt  ch P org respond) ;

: <ANALOG   ( -- )  \ Give 8-bit analog value
( ) ch A temit  8 4A *bis  adc-in 5 >pay  ch A respond ;

: +ABO      ( -- )      2 pay>  abo *set ;  \ Take abo
: -ABO      ( -- )      2 pay>  abo *clr ;  \ Free abo

\ Wireless network building routines
: COPY>     ( -- )      \ Receive ALL nodes table
    6 'pay>  all #map move ;

: COPYALL   ( -- )
    new >use  begin  use next? while
        all >payload  ch C swap >node
    repeat ;

: .DIRECT   ( -- )
    cr  ." Direct " direct .map ; \ Show all found direct nodes

: .NODES    ( -- )
   cr ." Nodes " nodes ccrange ?do  space  i c@ . i 1+ c@ .  2 +loop ;

: .TYPES    ( -- )      \ Show the all the found node types
    cr ." Node types: "  #n 0 ?do  i all get* if  types i + c@ .  then  loop ;

: CALLED    ( -- )      \ Code runned by external command!
    org? 0= if
        new count* dup t. 5 >pay \ Store nr. of new nodes
        new >payload    \ Send NEW nodes map too
        ch } org respond)
    then ;

\ Add a direct found node to my node tables
: NOTE-DIRECT   ( node -- )
    >r  r@ all get* 0= if \ Yes, node not recorded yet?
        r@ new *set \ Yes, note new node number!
    then
    r@ r@ nodes +cc \ Add direct node to table
    r@ direct *set  \ Save that node number
    r> all *set ;   \ Note node in ALL nodes table too

\ 0 scan = max. 4 meters  ( 1 muur )    2 scan = max. 10 meters ( 1 muur )
\ 4 scan = max. 10 meters ( 2 muren )   6 scan = max. .. meters ( .. )
: SCANX)    ( rf-setting -- ) \ Scan to other nodes within RF-range
    >nrf  new zero          \ Set RF-power, empty new node table
    nodes c@ 0= if          \ Not yet scanned?
        #me all *set        \ Yes, i am there too
        #n 0 ?do            \ Try all findable nodes
            i #me <> if     \ Do not scan yourself
                i set-dest
                ch S retry transceive 0= if  \ Node found?
                    tcr  i t.       \ Yes. show the scanned node number
                    t" Found" i note-direct
                then
            then
        loop
    then
    cr ." New: " new .map  nrf >nrf \ Copy ALL table & show
    copyall  .direct  called  read-mode ; \ show direct nodes found

: SCANX     ( -- )      4 scanx) ;  \ SCANX at medium power

: HOP)      ( c -- )        \ Add nested nodes to my tables
    to tk  direct >use                          \ Do all direct nodes
    begin  use next? while                      \ Noted node?
        tcr t" Node " dup t.  tk swap ask? if   \ Ask table
            0 pay> if
                begin  6 'pay> next? while      \ Node present?
                    dup direct get* 0= if       \ Yes, node not direct
                        dup t.  dup notdone? if \ Not indirect too
                            dup indirect *set   \ Add node
                            2 pay> over nodes +cc t" hop "
                        else  t" not "  then  dup all *set
                    then
                    drop
                repeat
            then
        then
    repeat  cr ." Hop to " indirect .map  called ;

: HOP       ( -- )      ch H hop) ;
: HOPP      ( -- )      ch N hop) ;

: GETINFO   ( -- )      \ Add node info to my type table
    all >use  tcr t" Info "
    begin  use next? while      \ Node exists?
       dup t.  dup #me = if     \ Myself?
            #type  over types + c! \ Yes, note my type
        else
            ch S over ask? if   \ No, ask info
                6 pay> over types + c! \ Received, add node-type
            then
        then
        drop
    repeat
    .types  .nodes  cr  called ;

\ Check node connection/availability
: PING      ( node -- )
    #me over <> if
        3 0 do
            cr base @ >r  decimal
            ." Node " dup .
            ch P over ask? if \ Ping 'node'
                ." count = " 5 pay> 6 pay> >< or
                0 <# # #s #> type  ."  time is "  tp
            else
                ." timeout after "  delay retry *
            then
            65 64 */  0 <# # ch . hold #s #> type  ."  ms "
            r> base !
        loop
    else  ." not on myself "  then  drop ;


\ Show stack depth as prompt
: .PROMPT   ( -- )
    depth 0 <#  ch > hold  # #  ch < hold  #> type  cr ;

\ Gather keypresses until <enter> is pressed
value CPTR
create 'BUF 18 allot    \ Input buffer for max. 24 chars
: FORTHCOMMAND  ( -- ?? )
    key? if
        key  dup emit  dup 8 = if \ Backspace?
            cptr if \ Line not empty
                space  dup emit  -1 +to cptr \ Delete char
             then  drop
        else
            dup 0D <> if  'buf cptr 17 umin + c!  incr cptr  exit  then
            drop  cr  'buf cptr 0 !input  interpret  0 to cptr
            read-mode  .prompt  ?stack  ( Stack check )
        then
    then ;

\ Handle commands & scripts
: NODE-ERROR    ( -- )
    t? if ."  Err " .prompt  then  -1 ?abort ; \ Signal error

\ Receive and execute commands from another node
value ON?       \ Switch On/Off
: OUTPUTON  ( -- )      -1 to on?  pwm 8 umax >power ;
: OUTPUTOFF ( -- )      0 to on?   00 >power ;
: POWER>    ( -- )      5 pay> 2* 2*  dup to pwm  >power ;
: FORTH>    ( -- ?? )   5 'pay>  0 pay>  0 !input  interpret  ch E respond ;


\ Node commands, because all nodes are equal each
\ node must be able to send and recieve commands.
\ This is done by selecting different modes.
create 'COMMANDS    ( -- addr )
  ( Execute generic script commands )
    ch F ,  ' forth> ,      \ Execute (Run) Forth command
  ( Execute target specific script commands )
    ch E ,  ' outputon ,    \ Activate (E)nable
    ch Q ,  ' outputoff ,   \ Deactivate (Q)uit
    ch % ,  ' power> ,      \ Set PWM for mosfet
    ch A ,  ' <analog ,     \ Read ADC input 3
    ch + ,  ' +abo ,        \ (Un)subscribe
    ch - ,  ' -abo ,
  ( Give back some network data )
    ch S ,  ' <scanx ,      \ Give nr. of direct nodes & info back
    ch H ,  ' <hop ,        \ Give my direct nodes back
    ch N ,  ' <hopp ,       \ Give my indirect nodes back
    ch P ,  ' <ping ,       \ Respond on ping
  ( Gather network data actively )
    ch C ,  ' copy> ,       \ Receive current ALL table
    ch s ,  ' scanx ,       \ Scan network
    ch i ,  ' getinfo ,     \ Gather node info
    ch h ,  ' hop ,         \ Ask direct nodes of neighbours
    ch n ,  ' hopp ,        \ Ask indirect nodes of neighbours
  ( Finish )
    ch } ,  ' ready ,       \ External command finished!
    ch | ,  ' halt ,        \ Stop any free running program
    -1 ,    ' node-error ,  \ Abort on an error

\ Wait MS milliseconds & respond to external network commands
: LISTEN    ( ms -- )
    >ms  read-mode  begin
        irq? if  xkey 'commands exec  then  (ms)
    end? until ;

\ Commands to other nodes directly
: ON        ( node -- ) ch E  swap >node  10 listen ; \ Send power on
: OFF       ( node -- ) ch Q  swap >node  10 listen ; \ Send power off
: STOP      ( node -- ) ch |  swap >node  10 listen ; \ Stop external Forth program

: SUBSCR    ( flag node -- )    \ Take (un)subscribtion at node
    swap if  ch +  else  ch -  then  swap >node  10 listen ; \ (Un)subscribe

: SUBSCRIBE ( a u -- )      \ Subscribe all nodes from string a u
    bounds ?do  i c@ ch 0 - abo *set  loop ;

: ALL-ON    ( -- )  all >use  begin  use next? while  on  repeat ;
: ALL-OFF   ( -- )  all >use  begin  use next? while  off  repeat ;
: STOPALL   ( -- )  all >use  begin  use next? while  stop  repeat ;

: SWITCH    ( -- )
    s? if  exit  then  begin  20 listen  s? until
    on? if  all-off  else  all-on  then  80 listen ;

value ADC?  \ ADC or switch event selected, default is switch  ADC? = 0
: HYST?     ( +n1 -- +n2 )      A0  on? if 10 + else 10 - then  < ;
: EVENT     ( -- )
    adc? if
        adc-in hyst? if  \ ADC active
            on? 0= if  all-on  then
        else  on? if  all-off  then  then
    else  switch  then ; \ Switch active

: HANDLER   ( -- )
    irq? if     \ Payload packet received?
        xkey            \ Yes, get it  \ c
        2 pay> to org   \ Remember origin node \ c
        1 pay> #me = if \ Packet for me? \ c
            'commands exec \ Yes execute, command \ -
        else
           ch S <> if   \ No SCANX ? \ -
                1 pay> >dest    \ Yes, fetch & set destination
                thru 4 pay> xemit \ Relay packet using it's own command
                t? if           \ Show hopping?
                    ."  Hop by " #me . ." to " 1 pay> .
                    ." dest = "  dest . cr
                then
            then
        then
        read-mode
    else                    \ No, own command
         ms? 0= if  event  forthcommand  then
    then
    (ms)  red-off ;

: NODE      ( -- )
    restart  pwm-setup  ms-on \ Tracer on, change to TROFF in RESTART
    key? if key drop  then  0 to cptr  restore-sp
    adc-setup  0 to adc?  read-mode  outputoff  cr
    begin
        ['] handler catch ?dup
        if  count 1F and type  ."  Error " \ Show error
            .prompt  restore-sp  free \ Init. datastack
            postpone [  0 to cptr \ Go executing
            setup24l01  read-mode \ Reinitialise on error
        then
    again ;


\ Send Forth command string of max. 0C bytes to remote node
: >F        ( node ccc -- )
    0 parse         \ Set destination, get command
    >r  r@ 0 >pay   \ String length to admin.
    5 '>pay  r> #b umin move \ Copy command to payload
    ch F swap >node ; \ Send command

: QUIT?     ( -- f )    key? org? or ;

: RUNNER    ( -- )      \ Running light on all outputs in my network
    begin
        all >use
        begin  use next? while
        dup on  100 listen
        off  30 listen  repeat
    quit? until  free ;

: FLASH1    ( -- )      \ Flash all outputs in my network
    begin  all-on  80 listen  all-off  300 listen  quit? until  free ;

: FLASH2    ( -- )      \ Double flash all outputs in my network
    begin
        all-on  20 listen  all-off  80 listen
        all-on  20 listen  all-off 300 listen
    quit? until  free ;

: FLASH3    ( -- )      \ Flash all outputs in my network
    begin  all-on  400 listen  all-off  1000 listen  quit? until  free ;

: ANALOG    ( node -- +n -1|0 ) \ Ask analog input from 'node'
    #me over = if  drop  adc-in  -1  exit  then
    ch A swap ask? if  5 pay>  -1  exit  then  0 ;

: .AD   ( node -- ) analog if  ." ADC:"  .  then ;
: AD    ( node -- ) begin  dup .ad  40 listen  quit? until  drop  free ;

: DIFF  ( -- ad ) \ Give new adc value but only of the difference is large enough
    adc-in
    begin  quit? 0= while
        10 listen  dup adc-in -
    abs 4 > until
    then  drop  adc-in ;

: DIM   ( map -- )  \ DIM all outputs given in the map from stack
    all-off                 \ m
    begin                   \ m
        dup >use  diff      \ m b
        begin  use next? while \ m b n
        over 5 >pay  ch % swap >node  10 listen \ m b
        repeat  drop        \ m
    quit? until  drop ;     \ -

: DIMMER ( -- )     abo dim  all-off ;


\ Build a network from scratch!
: NOTE-NEW  ( -- )  \ Note any new nested node
    begin  6 'pay> next? while  nested *set  repeat ;

: RUN       ( c node -- )   \ Execute command 'c' on node
    >node  start    \ Send command to node
    read-mode  begin  handler  end? until ; \ Respond until '}' was received

: OTHER     ( a c -- )      \ Perform command 'c' on all nodes from map a
    to tk  >use  #me use *clr \ Save token, copy table and
    tk ch s = if  nested zero  then \ When SCANX, clear new node table
    begin  use next? while
        tk swap run         \ Execute command and give response
        tk ch s = if        \ SCANX ?
            5 pay> if       \ New nodes?
                ."  new "  6 'pay> .map  note-new \ Yes, note
        then  then
    repeat ;

: BUILD     ( -- )
    nodes c@ if  reset  then    \ Nodes filled in, then empty all tables
    ." From " .status  halt     \ Show building node
    scanx  new  ch s other      \ SCANX on myself & all new found nodes
    hop   new  ch h other       \ HOP on myself & all new found nodes
\   Start SCANX on new found nested nodes etc.
    begin  nested count* while  \ New nodes found
        nested  ch s other      \ Do a SCANX there
        nested  ch h other      \ and a HOP
        all  ch n other         \ finally on all nodes a HOPP
    repeat
\   3 0 do  hopp  all  ch n other  loop \ Then HOPP is needed too
    getinfo  all  ch i other  free ; \ Gather all node information & ready


shield NODE\  troff \ Tracer off
' node  to app  freeze
